from typing import List

from loopgpt.tools import GoogleSearch, Browser
import loopgpt


@loopgpt.aifunc()
def get_keywords_and_questions(topic: str) -> List[str]:
    """This is an atomic function. This function helps the user learn about the given topic in detail.
    It returns a short list of 10 keywords and questions that will be optimal for searching the web.

    Args:
        topic (str): The topic to get keywords and questions for.

    Returns:
        List[str]: A list of keywords and questions about the topic.
    """


@loopgpt.aifunc(tools=[GoogleSearch, Browser])
def research(topic: str) -> List[str]:
    """This function searches the topic on Google and browses the first link. It returns a list of the most important sub-topics
    (in order of decreasing importance) obtained from the browsed pages that can be explored further.

    Args:
        topic (str): The topic to research.

    Returns:
        List[str]: A list of important sub-topics learned
    """


@loopgpt.aifunc()
def clean_up(index: str) -> str:
    """This is an atomic function. Remove duplicate sections and subsections from the given index.

    Args:
        index (str): Index to clean up.

    Returns:
        str: The cleaned up index.
    """


@loopgpt.aifunc()
def generate_section_and_paragraph_headings(context: str) -> List[str]:
    """This is an atomic function. It takes a context and generates a list of short and concise headings.

    Args:
        context (str): The context to generate headings for.

    Returns:
        List[str]: The list of headings generated.
    """

@loopgpt.aifunc()
def generate_index(bad_index: str) -> str:
    """This is a semantic function. Given is an index for a book, but it has many problems, for example, duplicate sections, 
    bad section names, subsections that don't match their sections, missing numbering or sections without subsections. 
    This function fixes all of these problems and returns a properly numbered good index without duplicate sections or missing subsections.

    Examples:
        1. Section 1
            1.1. Subsection 1
            1.2. Subsection 2
        2. Section 2
            3.1. Subsection 1
    
    Args:
        bad_index (str): The bad index to be fixed.

    Returns:
        str: The generated good index.
    """



from writers import *
from writers.latex import LATEX_TEMPLATE


def get_research_args(mode):
    if mode.endswith("short"):
        return {"breadth": 4, "depth": 1}
    else:
        return {"breadth": 4, "depth": 2}


def get_writer(mode):
    if mode == "long":
        return TextWriter
    elif mode == "short":
        return ShortTextWriter
    elif mode == "latex-long":
        return LatexWriter
    elif mode == "latex-short":
        return ShortLatexWriter
    else:
        raise ValueError("Invalid mode.")

def get_template(mode):
    if mode.startswith("latex"):
        return LATEX_TEMPLATE
    else:
        return "{content}"



import loopgpt
import openai
import os

# openai.api_type = "azure"
# openai.api_base = "https://loopgpt-azure-openai.openai.azure.com/"
# openai.api_version = "2023-03-15-preview"
# openai.api_key = os.getenv("AZURE_OPENAI_KEY")

from loopgpt.models import AzureOpenAIModel, OpenAIModel
from loopgpt.embeddings import AzureOpenAIEmbeddingProvider, OpenAIEmbeddingProvider

# model = AzureOpenAIModel("loop-gpt-35-turbo")
# emb = AzureOpenAIEmbeddingProvider("loop-text-embedding-ada-002")
model = OpenAIModel()
emb = OpenAIEmbeddingProvider()

loopgpt.set_aifunc_args(model=model, embedding_provider=emb)



import loopgpt

from abc import ABC, abstractmethod
from functools import wraps
from typing import Callable


class Writer(ABC):
    @staticmethod
    @abstractmethod
    @loopgpt.aifunc()
    def write_section(section: str) -> str:
        ...

    @staticmethod
    @abstractmethod
    @loopgpt.aifunc()
    def write_subsection(subsection: str) -> str:
        ...


class format:
    def __init__(self, format: Callable):
        self.format = format

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            response = func(*args, **kwargs)
            return self.format(*args, response, **kwargs)

        return inner



from collections import defaultdict, OrderedDict

from utils import (
    get_keywords_and_questions,
    generate_index,
)
from loopgpt.tools import GoogleSearch, Browser
from modes import get_writer, get_template
from datetime import date


def research(topic, research_agent, breadth=3, depth=1):
    keywords_and_questions = get_keywords_and_questions(topic)[:breadth]

    google_search = GoogleSearch()
    browser = Browser()

    data_sources = {}
    keywords_map = defaultdict(list)

    i = 0
    keywords_map[i] = keywords_and_questions[:]
    with research_agent:
        while depth > 0:
            new_keywords = []
            while keywords_and_questions:
                search_term = keywords_and_questions.pop(0)
                _, links = google_search.manual_run(topic + " " + search_term)
                for j in range(breadth):
                    if data_sources.get(links[j]):
                        pass
                    else:
                        data_sources[links[j]] = browser.run(links[j], "")

                with research_agent.query(search_term):
                    subtopics = get_keywords_and_questions(search_term)
                    next_level = subtopics[:4]
                    next_keywords = subtopics[:breadth]

                keywords_map[i + 1].append(next_level[:])
                new_keywords.extend(next_keywords[:])
            keywords_and_questions = new_keywords[:]
            depth -= 1
            i += 1
    
    keywords_map.pop(i)

    heading_map = ""

    def print_section(depth, section, lst):
        nonlocal heading_map
        heading_map += f"{'    ' * depth}{lst[section]}\n"
        if keywords_map[depth + 1]:
            for subsection in range(len(keywords_map[depth + 1][section])):
                print_section(depth + 1, subsection, keywords_map[depth + 1][section])

    for section in range(len(keywords_map[0])):
        print_section(0, section, keywords_map[0])

    index = generate_index(heading_map)

    return index


def write_book(topic, index, writer_agent, filename, mode):
    items = index.split("\n")

    file = open(filename, "w")

    current_heading = ""

    writer = get_writer(mode)
    template = get_template(mode)

    content = ""
    last_section = ""
    with writer_agent:
        for item in items:
            n, heading = item.strip().split(" ", 1)
            if len(n) == 2:
                with writer_agent.query(heading):
                    last_section = writer.write_section(heading)
                    content += last_section + "\n\n"
                current_heading = heading
            else:
                with writer_agent.complete({"assistant": last_section}):
                    with writer_agent.query(current_heading + ": " + heading):
                        content += writer.write_subsection(heading) + "\n\n"
    file.write(template.format(title=topic, date=str(date.today()), content=content))
    file.close()



from .base import Writer, format

import loopgpt

TEXT_TEMPLATE = """{title}

{date}

{content}
"""


class TextWriter(Writer):
    @staticmethod
    @format(
        lambda heading, content: f"\n{heading}\n{'=' * len(heading)}\n\n{content}"
    )
    @loopgpt.aifunc()
    def write_section(section: str) -> str:
        """Writes a short section about the given topic without going into specific details as there will be future subsections for that.
        The section should be less than 50 words long and should mention at least 3 subtopics that will be covered in future subsections.

        Args:
            section (str): The topic of the section.
        
        Returns:
            str: The section content.
        """
    
    @staticmethod
    @format(
        lambda heading, content: f"\n{heading}\n{'-' * len(heading)}\n\n{content}"
    )
    @loopgpt.aifunc()
    def write_subsection(subsection: str) -> str:
        """Writes a detailed subsection about the given topic that should contain as much details as possible including reference links.
        The subsection should be between 500 and 1000 words long and should cover the topic in depth. Please include at least 1 reference link 
        from a reputable source.

        Args:
            subsection (str): The topic of the subsection.
        
        Returns:
            str: The subsection content.
        """


class ShortTextWriter(TextWriter):
    @staticmethod
    @loopgpt.aifunc()
    @format(
        lambda heading, content: f"\n{heading}\n{'-' * len(heading)}\n\n{content}"
    )
    def write_subsection(subsection: str) -> str:
        """Writes a short subsection about the given topic that should be between 100 and 200 words long. Inlcudes any relevant links that were found 
        previously where the reader can find more information.

        Args:
            subsection (str): The topic of the subsection.

        Returns:
            str: The subsection content.
        """



from .base import Writer

import loopgpt

LATEX_TEMPLATE = """\\documentclass{{report}}

\\title{{{title}}}
\\date{{{date}}}

\\begin{{document}}

\\maketitle

\\tableofcontents

{content}

\\end{{document}}
"""


class LatexWriter(Writer):
    @staticmethod
    @loopgpt.aifunc()
    def write_section(section: str) -> str:
        """Writes a latex \section{} command with the section name and then a short section about the given topic without going into specific details 
        as there will be future subsections for that. The section should be less than 50 words long and should mention at least 3 subtopics 
        that will be covered in future subsections.

        Args:
            section (str): The section to write the introduction for.

        Returns:
            str: The introduction written in LaTeX format.
        """

    @staticmethod
    @loopgpt.aifunc()
    def write_subsection(subsection: str) -> str:
        """Writes a latex \subsection{} command with the subsection name and then a detailed subsection about the given topic that should contain as much 
        details as possible including reference links. The subsection should be between 500 and 1000 words long and should cover the topic in depth. 
        It must be in valid LaTeX with correct escaping.

        Args:
            subsection (str): The subsection to write about.

        Returns:
            str: The section written in LaTeX format.
        """


class ShortLatexWriter(LatexWriter):
    @staticmethod
    @loopgpt.aifunc()
    def write_subsection(subsection: str) -> str:
        """This is a semantic function. It writes a latex subsection command followed by the subsection name and then a very short subsection
        content including links to relevant websites.
        Writes a latex \subsection{} command with the subsection name and then a short subsection about the given topic that should be between 100 and 200 
        words long. It must be in valid LaTeX with correct escaping. Includes any relevant links that were found previously where the reader can 
        find more information.

        Args:
            subsection (str): The subsection to write about.

        Returns:
            str: The section written in LaTeX format.
        """



from setuptools import setup, find_packages

package_name = "researchgpt"

install_requires = ["loopgpt"]
extras_require = {}

if __name__ == "__main__":
    setup(
        install_requires=install_requires,
        extras_require=extras_require,
        packages=find_packages(),
        name=package_name,
        version="0.0.1",
        description="Modular Auto-GPT Framework",
        entry_points={"console_scripts": ["loopgpt = loopgpt.loops.cli:main"]},
    )



import argparse
import json

from loopgpt.aifunc import create_empty_agent
from main import research, write_book
from modes import get_research_args
from config import model, emb


research_agent = create_empty_agent(model=model, embedding_provider=emb)
writer_agent = create_empty_agent(model=model, embedding_provider=emb)

parser = argparse.ArgumentParser()

parser.add_argument("topic", help="Topic to research")
parser.add_argument(
    "filename",
    help="Filename to write the research.",
)
parser.add_argument(
    "--mode",
    choices=["short", "long"],
    help="Mode of research.",
    default="short",
)
parser.add_argument(
    "--latex",
    help="Latex mode",
    action="store_true",
)
parser.add_argument(
    "--breadth",
    help="Breadth of research.",
)
parser.add_argument(
    "--depth",
    help="Depth of research.",
)

args = parser.parse_args()

topic = args.topic
filename = args.filename
mode = args.mode

if args.latex:
    mode = "latex-" + mode


research_agent = create_empty_agent(model=model, embedding_provider=emb)
writer_agent = create_empty_agent(model=model, embedding_provider=emb)

writer_agent.memory = research_agent.memory

research_args = get_research_args(mode)
if args.breadth:
    research_args["breadth"] = int(args.breadth)
if args.depth:
    research_args["depth"] = int(args.depth)

index = research(topic, research_agent, **research_args)
write_book(topic, index, writer_agent, filename, mode)
